;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "grip-mode" "20260610.628"
  "Instant GitHub-flavored Markdown/Org preview using grip."
  '((emacs "24.4"))
  :url "https://github.com/seagle0128/grip-mode"
  :commit "6ed3f9739e4a3320ab063f923c00215f7ee5b5cc"
  :revdesc "6ed3f9739e4a"
  :keywords '("convenience" "markdown" "preview")
  :authors '(("Vincent Zhang" . "seagle0128@gmail.com"))
  :maintainers '(("Vincent Zhang" . "seagle0128@gmail.com")))
